///////////////////////////////////////////////////////////////////////////////
// Dot3 env.
//
// 2017.04.21.()
//-----------------------------------------------------------------------------
// 
// 1. Microsoft DirectX SDK (February 2010), Tut06_Meshes Ʈ ⺻
// 2. vs 6.0 ҽ vs2013 ȯ http://www.dimmension3.spine.pl/modules.php?name=Tutorials&wtd=show_cat&cat=direct3d 
//-----------------------------------------------------------------------------

#pragma comment(lib, "dxerr.lib")

#include <Windows.h>
#include <mmsystem.h>
#include <d3dx9.h>
#pragma warning( disable : 4996 ) // disable deprecated warning 
#include <tchar.h>
#include <strsafe.h>
#pragma warning( default : 4996 )
#include <DxErr.h>

#include <assert.h>
#include <fstream>				// std::ofstream
#include "resource.h"
#include "DialogProc.h"
#include "MeshUtils.h"


//-----------------------------------------------------------------------------
// Miscellaneous helper functions
//-----------------------------------------------------------------------------
#define SAFE_DELETE(p)       { if(p) { delete (p);     (p)=NULL; } }
#define SAFE_DELETE_ARRAY(p) { if(p) { delete[] (p);   (p)=NULL; } }
#define SAFE_RELEASE(p)      { if(p) { (p)->Release(); (p)=NULL; } }


//=============================================================================
// all sample objects
//-----------------------------------------------------------------------------
const char * file_path = "Media";

#define	SAMPLE_VS11			"dot3env_shader.vsh"
#define	SAMPLE_VS20			"dot3env_shader_vs.2.0.vsh"
#define	SAMPLE_PS11			"dot3env_shader.psh"
#define	SAMPLE_PS14			"dot3env_shader_ps.1.4.psh"
#define	SAMPLE_PS20			"dot3env_shader_ps.2.0.psh"

#define EARTH_MESH			"Media/earth.x"
#define EARTH_NORMAL_TEX	"Media/earth_normal.tga"
#define WHITE_TEX			"Media/white.tga"
#define ENV_CUBE_MAP		"Media/LobbyCube.dds"

// main variables
HWND					g_hWnd				= NULL; // main rendering window
RECT					g_rect; 					// window sizes
LPDIRECT3D9				g_lpD3D				= NULL; // Used to create the D3DDevice
LPDIRECT3DDEVICE9		g_lpD3DDevice		= NULL; // Our rendering device
LPDIRECT3DVERTEXSHADER9 g_lpVertexShader_11	= NULL;
LPDIRECT3DVERTEXSHADER9 g_lpVertexShader_20 = NULL;
LPDIRECT3DPIXELSHADER9	g_lpPixelShader_11	= NULL;
LPDIRECT3DPIXELSHADER9	g_lpPixelShader_14	= NULL;
LPDIRECT3DPIXELSHADER9	g_lpPixelShader_20	= NULL;
D3DPRESENT_PARAMETERS	g_D3Dpp;					// present parameters structure
D3DCAPS9				m_d3dCaps;					// Caps for the device 

// mesh
LPD3DXMESH				g_lpMesh;
DWORD					g_dwMeshMaterials;
LPDIRECT3DTEXTURE9*		g_lpMeshTextures;
LPDIRECT3DTEXTURE9		g_lpMeshNormalTexture;
LPDIRECT3DTEXTURE9		g_lpWhiteTexture;
LPDIRECT3DCUBETEXTURE9	g_lpCubeEnvTexture;

// other variables
char				szPath[1024];				// path to media files
FLOAT				g_fAspect;
BOOL				m_bPS14Available;

//Our font interface
ID3DXFont *				g_font = NULL;


// A structure for our custom vertex type
struct SVertex
{
	float		x, y, z;		// position
	float		nx, ny, nz;		// normal
	float		u1, v1;			// firts texture channel
	float		ux, uy, uz;		// U vector
	float		vx, vy, vz;		// V vector
	float		uvx, uvy, uvz;	// UxV vector
};

// Our custom FVF, which describes our custom vertex structure
#define FVF_VERTEX ( D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX4 | \
	D3DFVF_TEXCOORDSIZE2(0) | D3DFVF_TEXCOORDSIZE3(1) | \
	D3DFVF_TEXCOORDSIZE3(2) | D3DFVF_TEXCOORDSIZE3(3))

// struct used to draw simple vectors
struct SVertexVector
{
	SVector3f	pos;	// position
	DWORD		color;	// color
};

#define FVF_VERTEXVECTOR ( D3DFVF_XYZ | D3DFVF_DIFFUSE )

// control program
enum EMode
{
	eNormalMap,		// normal map applied to object
	eDiffuseMap,	// diffuse map applied to object
	eEnvBump,		// enviroment map modified by normal
	eFinal,			// enviroment map modified by normal + diffuse texture
};

EMode		g_eMode = eFinal;
bool		g_bWireframe = false;
bool		g_bNormals = false;

//-----------------------------------------------------------------------------
enum EVertexShader
{
	eVS_None,
	eVS_1_1,		
	eVS_2_0,
};

int g_eVS_Start = eVS_None;
int g_eVS		= eVS_None;
int	g_eVS_MAX	= eVS_None;

enum EPixelShader
{
	ePS_None,
	ePS_1_1,
	ePS_1_4,
	ePS_2_0,
};

int	g_ePS_Start = ePS_None;
int g_ePS		= ePS_None;
int	g_ePS_MAX	= ePS_None;

//-----------------------------------------------------------------------------
// buffers to draw tangent space
int						g_nNormals = 0;
SVertexVector*			g_pNormals = NULL;
SVertexVector*			g_pTangents = NULL;
SVertexVector*			g_pBinormals = NULL;

//=============================================================================
// functions declarations
//-----------------------------------------------------------------------------
LRESULT WINAPI MsgProc(HWND, UINT, WPARAM, LPARAM);
INT WINAPI wWinMain(HINSTANCE hInst, HINSTANCE, LPWSTR, INT);
bool Initialize(HWND hWnd);
void DeInitialize();
void Render();
void Reset();
int ProcessKey(unsigned int);
int ProcessCommand(unsigned int);
bool InitGeometry();
void FreeGeometry();

bool CreateVertexShader();
	HRESULT CreateVS(const char * strVSName, LPDIRECT3DVERTEXSHADER9 * pVertexShader);
void FreeVertexShader();
void ChangeVertexShader();

bool CreatePixelShader();
	HRESULT CreatePS(const char * strPSName, LPDIRECT3DPIXELSHADER9 * pPixelShader);
void FreePixelShader();
void ChangePixelShader();


//=============================================================================
// Name: MsgProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------
LRESULT WINAPI MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_SIZE:
	{
		if (NULL != g_lpD3DDevice)
		{
			Reset();
		}
	}
	break;

	case WM_DESTROY:
	{
		DeInitialize();
		PostQuitMessage(0);
	}
	return 0;

	case WM_KEYDOWN:
	{
		ProcessKey((int)wParam);
	}
	break;

	case WM_COMMAND:
	{
		ProcessCommand(LOWORD(wParam));
	}
	break;

	default:
		break;
	}

	return DefWindowProc(hWnd, msg, wParam, lParam);
}

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: The application's entry point
//-----------------------------------------------------------------------------
INT WINAPI wWinMain(HINSTANCE hInst, HINSTANCE, LPWSTR, INT)
{
	WNDCLASS	wndClass;

	wndClass.style = CS_HREDRAW | CS_VREDRAW;
	wndClass.lpfnWndProc = MsgProc;
	wndClass.cbClsExtra = 0L;
	wndClass.cbWndExtra = 0L;
	wndClass.hInstance = GetModuleHandle(NULL);
	wndClass.hIcon = NULL;
	wndClass.hCursor = NULL;
	wndClass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	wndClass.lpszMenuName = MAKEINTRESOURCE(IDR_MENU);
	wndClass.lpszClassName = "D3D Tutorial";

	// register window class
	if (0 == RegisterClass(&wndClass))
	{
		MessageBox(NULL, "Cannot register window class", "Error", MB_OK | MB_ICONEXCLAMATION);

		return -1;
	}

	// create window - center on screen
	int		nX = GetSystemMetrics(SM_CXSCREEN);
	int		nY = GetSystemMetrics(SM_CYSCREEN);

	nX = nX / 2 - 300;
	nY = nY / 2 - 300;
	g_hWnd = CreateWindow("D3D Tutorial", "D3D Tutorial - Dot3 env.",
		WS_OVERLAPPEDWINDOW, nX, nY, 600, 600,
		GetDesktopWindow(), NULL, wndClass.hInstance, NULL);

	if (NULL == g_hWnd)
	{
		MessageBox(NULL, "Cannot create window", "Error", MB_OK | MB_ICONEXCLAMATION);
		UnregisterClass("D3D Tutorial", wndClass.hInstance);

		return -1;
	}

	// Show the window
	ShowWindow(g_hWnd, SW_SHOWDEFAULT);
	UpdateWindow(g_hWnd);

	if (true == Initialize(g_hWnd))
	{
		// Enter the message loop
		MSG msg;
		ZeroMemory(&msg, sizeof(msg));

		while (msg.message != WM_QUIT)
		{
			if (PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
			else
			{
				Render();
			}
		}
	}
	else
	{
		MessageBox(NULL, "Cannot initialize renderer", "Error", MB_OK | MB_ICONEXCLAMATION);
	}

	UnregisterClass("D3D Tutorial", wndClass.hInstance);

	return 0;
}

//-----------------------------------------------------------------------------
// Name: Initialize
// Desc: Initializes Direct3D
//-----------------------------------------------------------------------------
bool Initialize(HWND hWnd)
{
	HRESULT 				hResult = 0;
	D3DDISPLAYMODE			D3Ddm;

	// store main window
	assert(hWnd);
	g_hWnd = hWnd;

	// Create the D3D object.
	if (NULL == (g_lpD3D = Direct3DCreate9(D3D_SDK_VERSION)))
		return false;
		
	// get window rect
	GetClientRect(g_hWnd, &g_rect);
	g_fAspect = (float)(g_rect.right - g_rect.left) / (float)(g_rect.bottom - g_rect.top);

	// get current display mode
	hResult = g_lpD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &D3Ddm);

	if (D3D_OK != hResult)
	{
		return false;
	}


	// Set up the structure used to create the D3DDevice. Since we are now
	// using more complex geometry, we will create a device with a zbuffer.
	D3DPRESENT_PARAMETERS d3dpp;
	ZeroMemory(&d3dpp, sizeof(d3dpp));
	d3dpp.Windowed = TRUE;
	d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
	d3dpp.EnableAutoDepthStencil = TRUE;
	d3dpp.AutoDepthStencilFormat = D3DFMT_D16;

	// Create the D3DDevice
	if (FAILED(g_lpD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
		D3DCREATE_SOFTWARE_VERTEXPROCESSING,
		&d3dpp, &g_lpD3DDevice)))
	{
		return false;
	}

	g_lpD3DDevice->GetDeviceCaps(&m_d3dCaps);

	g_lpD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
	g_lpD3DDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
	g_lpD3DDevice->SetRenderState(D3DRS_ZENABLE, TRUE);
			
	if (false == InitGeometry())
	{
		return false;
	}

	if (false == CreateVertexShader())
	{
		return false;
	}

	if (false == CreatePixelShader())
	{
		return false;
	}

	hResult = D3DXCreateFont(	g_lpD3DDevice,     //D3D Device
								16,               //Font height
								0,                //Font width
								FW_NORMAL,        //Font Weight
								1,                //MipLevels
								false,            //Italic
								DEFAULT_CHARSET,  //CharSet
								OUT_DEFAULT_PRECIS, //OutputPrecision
								ANTIALIASED_QUALITY, //Quality
								DEFAULT_PITCH | FF_DONTCARE,//PitchAndFamily
								"Arial",			//pFacename,
								&g_font);			//ppFont

	return true;
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void DeInitialize()
{
	//Free up some internal resources
	g_font->OnLostDevice();
	
	if (g_font){
		g_font->Release();
		g_font = NULL;
	}

	// geometry
	FreeGeometry();
	FreeVertexShader();
	FreePixelShader();

	// render objects
	SAFE_RELEASE(g_lpD3D);
	SAFE_RELEASE(g_lpD3DDevice);
}

//-----------------------------------------------------------------------------
// Name: SetupMatrices()
// Desc: Sets up the world, view, and projection transform matrices.
//-----------------------------------------------------------------------------
VOID SetupMatrices()
{
	D3DXMATRIX		matWorld;
	D3DXMATRIX		matView;
	D3DXMATRIX		matProj;
	D3DXMATRIX		matViewProj;
	D3DXVECTOR4		vectEye;

	// world transform
	D3DXMatrixRotationY(&matWorld, timeGetTime() / 1500.0f);

	// view transform
	D3DXMatrixLookAtLH(&matView, &D3DXVECTOR3(0.0f, 2.0f, -6.0f),
		&D3DXVECTOR3(0.0f, 0.0f, 0.0f),
		&D3DXVECTOR3(0.0f, 1.0f, 0.0f));
	// projection transform
	D3DXMatrixPerspectiveFovLH(&matProj, D3DX_PI / 4, g_fAspect, 1.0f, 100.0f);

	// transform mode
	g_lpD3DDevice->SetTransform(D3DTS_WORLD, &matWorld);
	g_lpD3DDevice->SetTransform(D3DTS_VIEW, &matView);
	g_lpD3DDevice->SetTransform(D3DTS_PROJECTION, &matProj);

	// shaders data
	matViewProj = matView * matProj;
	vectEye = D3DXVECTOR4(0.0f, 0.0f, -6.0f, 0.0f);
	g_lpD3DDevice->SetVertexShaderConstantF(0, (float*)D3DXMatrixTranspose(&matWorld, &matWorld), 4);
	g_lpD3DDevice->SetVertexShaderConstantF(4, (float*)D3DXMatrixTranspose(&matViewProj, &matViewProj), 4);
	g_lpD3DDevice->SetVertexShaderConstantF(8, (float*)&vectEye, 1);
}

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Draws the scene
//-----------------------------------------------------------------------------
void Render()
{
	g_lpD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0, 0, 255), 1.0f, 0);
	g_lpD3DDevice->BeginScene();
	{
		// setup matrices
		SetupMatrices();

		// toggle wireframe
		if (true == g_bWireframe)
		{
			g_lpD3DDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);
		}
		else
		{
			g_lpD3DDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
		}

		// texture filtering
		g_lpD3DDevice->SetFVF(FVF_VERTEX);
		g_lpD3DDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
		g_lpD3DDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
		g_lpD3DDevice->SetSamplerState(3, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
		g_lpD3DDevice->SetSamplerState(3, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);

		if (eVS_1_1 == g_eVS)
			g_lpD3DDevice->SetVertexShader(g_lpVertexShader_11);
	
		if (eVS_2_0 == g_eVS)
			g_lpD3DDevice->SetVertexShader(g_lpVertexShader_20);
			
		if (ePS_1_1 == g_ePS)
			g_lpD3DDevice->SetPixelShader(g_lpPixelShader_11);

		if (ePS_1_4 == g_ePS)
			g_lpD3DDevice->SetPixelShader(g_lpPixelShader_14);

		if (ePS_2_0 == g_ePS)
			g_lpD3DDevice->SetPixelShader(g_lpPixelShader_20);


		// render different phases
		switch (g_eMode)
		{
		case eNormalMap:
		{
						   // shaders settings
						   g_lpD3DDevice->SetPixelShader(NULL);
						   // g_lpD3DDevice->SetVertexShader(g_lpVertexShader_11);

						   // texture settings - pass normal map to final stage
						   g_lpD3DDevice->SetTexture(0, g_lpMeshNormalTexture);
						   g_lpD3DDevice->SetTexture(1, NULL);
						   g_lpD3DDevice->SetTexture(2, NULL);
						   g_lpD3DDevice->SetTexture(3, NULL);

						   g_lpD3DDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG2);
						   g_lpD3DDevice->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_TEXTURE);

						   g_lpD3DDevice->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
						   g_lpD3DDevice->SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_CURRENT);

						   g_lpD3DDevice->SetTextureStageState(2, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
						   g_lpD3DDevice->SetTextureStageState(2, D3DTSS_COLORARG1, D3DTA_CURRENT);

						   g_lpD3DDevice->SetTextureStageState(3, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
						   g_lpD3DDevice->SetTextureStageState(3, D3DTSS_COLORARG1, D3DTA_CURRENT);

						   // draw mesh
						   for (DWORD i = 0; i < g_dwMeshMaterials; i++)
						   {
							   g_lpMesh->DrawSubset(i);
						   }
		}
			break;

		case eDiffuseMap:
		{
							g_lpD3DDevice->SetPixelShader(NULL);
							//g_lpD3DDevice->SetVertexShader(g_lpVertexShader_11);

							// texture settings - pass diffuse map to final stage
							g_lpD3DDevice->SetTexture(1, NULL);
							g_lpD3DDevice->SetTexture(2, NULL);
							g_lpD3DDevice->SetTexture(3, NULL);

							// draw mesh
							for (DWORD i = 0; i < g_dwMeshMaterials; i++)
							{
								g_lpD3DDevice->SetTexture(0, g_lpMeshTextures[i]);
								g_lpMesh->DrawSubset(i);
							}

							g_lpD3DDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG2);
							g_lpD3DDevice->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_TEXTURE);

							g_lpD3DDevice->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
							g_lpD3DDevice->SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_CURRENT);

							g_lpD3DDevice->SetTextureStageState(2, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
							g_lpD3DDevice->SetTextureStageState(2, D3DTSS_COLORARG1, D3DTA_CURRENT);

							g_lpD3DDevice->SetTextureStageState(3, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
							g_lpD3DDevice->SetTextureStageState(3, D3DTSS_COLORARG1, D3DTA_CURRENT);
		}
			break;

		case eEnvBump:
		{
						 // shaders settings
						 //g_lpD3DDevice->SetPixelShader(g_lpPixelShader_11);
						 //g_lpD3DDevice->SetVertexShader(g_lpVertexShader_11);

						 // texture settings - pass normal map to final stage
						 g_lpD3DDevice->SetTexture(0, g_lpMeshNormalTexture);
						 g_lpD3DDevice->SetTexture(1, NULL);
						 g_lpD3DDevice->SetTexture(2, NULL);
						 g_lpD3DDevice->SetTexture(3, g_lpCubeEnvTexture);

						 // draw mesh
						 for (DWORD i = 0; i < g_dwMeshMaterials; i++)
						 {
							 g_lpMesh->DrawSubset(i);
						 }

		}
			break;

		case eFinal:
		{
					   // shaders settings
					   //g_lpD3DDevice->SetPixelShader(g_lpPixelShader_11);
					   //g_lpD3DDevice->SetVertexShader(g_lpVertexShader_11);

					   // texture settings - pass normal map to final stage
					   g_lpD3DDevice->SetTexture(0, g_lpMeshNormalTexture);
					   g_lpD3DDevice->SetTexture(1, NULL);
					   g_lpD3DDevice->SetTexture(2, NULL);
					   g_lpD3DDevice->SetTexture(3, g_lpCubeEnvTexture);

					   // draw mesh
					   DWORD i;
					   for ( i = 0; i < g_dwMeshMaterials; i++)
					   {
						   g_lpMesh->DrawSubset(i);
					   }

					   // set appropriate render states
					   g_lpD3DDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
					   g_lpD3DDevice->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_TEXTURE);
					   g_lpD3DDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_DIFFUSE);

					   g_lpD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
					   g_lpD3DDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_SRCCOLOR);
					   g_lpD3DDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_DESTCOLOR);
					   g_lpD3DDevice->SetPixelShader(NULL);
					   g_lpD3DDevice->SetTexture(1, NULL);
					   g_lpD3DDevice->SetTexture(2, NULL);
					   g_lpD3DDevice->SetTexture(3, NULL);

					   // draw mesh
					   for (i = 0; i < g_dwMeshMaterials; i++)
					   {
						   g_lpD3DDevice->SetTexture(0, g_lpMeshTextures[i]);
						   g_lpMesh->DrawSubset(i);
					   }

					   // restore render states
					   g_lpD3DDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
		}
			break;
		}

		// draw normals
		if (true == g_bNormals)
		{
			g_lpD3DDevice->SetVertexShader(NULL);
			g_lpD3DDevice->SetPixelShader(NULL);
			g_lpD3DDevice->SetFVF(FVF_VERTEXVECTOR);
			g_lpD3DDevice->SetTexture(0, NULL);
			g_lpD3DDevice->SetTexture(1, NULL);
			g_lpD3DDevice->SetTexture(2, NULL);
			g_lpD3DDevice->SetTexture(3, NULL);
			g_lpD3DDevice->DrawPrimitiveUP(D3DPT_LINELIST, g_nNormals, g_pNormals, sizeof(SVertexVector));
			g_lpD3DDevice->DrawPrimitiveUP(D3DPT_LINELIST, g_nNormals, g_pTangents, sizeof(SVertexVector));
			g_lpD3DDevice->DrawPrimitiveUP(D3DPT_LINELIST, g_nNormals, g_pBinormals, sizeof(SVertexVector));
		}
	}

	RECT rect;
	SetRect(&rect, 10, 10, g_rect.right-10, g_rect.bottom-10);

	char szVS[64] = "";

	if ( eVS_1_1 == g_eVS)
		sprintf_s(szVS, "VS_1_1");

	if ( eVS_2_0 == g_eVS)
		sprintf_s(szVS, "VS_2_0");

	char szPS[64] = "";

	if ( ePS_1_1 == g_ePS)
		sprintf_s(szPS, "PS_1_1");

	if ( ePS_1_4 == g_ePS)
		sprintf_s(szPS, "PS_1_4");

	if ( ePS_2_0 == g_ePS)
		sprintf_s(szPS, "PS_2_0");

	char szTxt[256]; 
	sprintf_s(szTxt, "F7 ChageVertexShader ; %s \nF8 ChandePixelShader ; %s", szVS, szPS);

	if ("" == szVS)
	{
		OutputDebugString("?? Check ; "" == szVS \n");
	}
		
	int font_height = g_font->DrawText(NULL,        //pSprite
				szTxt,								//pString
				-1,									//Count
				&rect,								//pRect
				DT_LEFT | DT_NOCLIP,				//Format,
				0xFFFFFFFF);						//Color
	if (!font_height){
			OutputDebugString("Error writing to log file\n");
	}

	g_lpD3DDevice->EndScene();
	g_lpD3DDevice->Present(NULL, NULL, NULL, NULL);
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void Reset()
{
	HRESULT		hResult;

	// free resources
	FreeGeometry();
	FreeVertexShader();
	FreePixelShader();

	// update window data
	assert(g_hWnd);
	GetClientRect(g_hWnd, &g_rect);
	g_fAspect = (float)(g_rect.right - g_rect.left) / (float)(g_rect.bottom - g_rect.top);

	// reset device
	g_D3Dpp.BackBufferWidth = g_rect.right - g_rect.left;
	g_D3Dpp.BackBufferHeight = g_rect.bottom - g_rect.top;
	hResult = g_lpD3DDevice->Reset(&g_D3Dpp);

	if (D3D_OK != hResult)
	{
		return;
	}

	// device settings
	g_lpD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
	g_lpD3DDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
	g_lpD3DDevice->SetRenderState(D3DRS_ZENABLE, TRUE);

	// load resource again
	InitGeometry();
	if (false == CreateVertexShader())
	{
		return;
	}

	if (false == CreatePixelShader())
	{
		return;
	}
}

//=============================================================================
//-----------------------------------------------------------------------------
void ChangeVertexShader()
{
	if (eVS_None < g_eVS)
	{
		if (g_eVS_MAX == g_eVS)
		{
			g_eVS = g_eVS_Start;
		}
		else
		{
			g_eVS = g_eVS + 1;
		}
	}
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void ChangePixelShader()
{
	if (ePS_None < g_ePS)
	{
		if (g_ePS_MAX == g_ePS)
		{
			g_ePS = g_ePS_Start;
		}
		else
		{
			g_ePS = g_ePS + 1;
		}
	}
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
int ProcessKey(unsigned int nKey)
{
	switch (nKey)
	{
		case VK_ESCAPE:
			DeInitialize();
			PostQuitMessage(0);
			break;

		case 0x48:
		{
			DialogBox(GetModuleHandle((LPCTSTR) "d3d_dot3env.exe"), MAKEINTRESOURCE(IDD_HELP_DIALOG), g_hWnd, DialogProc);
		}
		break;

		// 'w'
		case 0x57:
		{
			g_bWireframe = !g_bWireframe;
		}
		break;

		// F1
		case 0x70:
		{
			g_eMode = eNormalMap;
		}
		break;

		// F2
		case 0x71:
		{
			g_eMode = eDiffuseMap;
		}
		break;

		// F3
		case 0x72:
		{
			g_eMode = eEnvBump;
		}
		break;

		// F4
		case 0x73:
		{
			g_eMode = eFinal;
		}
		break;

		// F5
		case 0x74:
		{
			g_bNormals = !g_bNormals;
		}
		break;

		// F7
		case 0x76:
		{
			ChangeVertexShader();
		}
		break;

		// F8
		case 0x77:
		{
			ChangePixelShader();
		}
		break;

		default:
		break;
	}

	return 0;
}

int ProcessCommand(unsigned int nMessage)
{
	switch (nMessage)
	{
	case ID_SHOW_HELP:
	{
						 DialogBox(GetModuleHandle((LPCTSTR) "d3d_dot3env.exe"), MAKEINTRESOURCE(IDD_HELP_DIALOG), g_hWnd, DialogProc);
	}
		break;

		// W
	case ID_DOT3_WIREFRAME:
		g_bWireframe = !g_bWireframe;
		break;

		// F2
	case ID_DOT3_NORMAL_TEX:
		g_eMode = eNormalMap;
		break;

		// F3
	case ID_DOT3_DIFFUSE_TEX:
		g_eMode = eDiffuseMap;
		break;

		// F4
	case ID_DOT3_BUMP:
		g_eMode = eEnvBump;
		break;

		// F5
	case ID_DOT3_BUMP_DIFFUSE:
		g_eMode = eFinal;
		break;

	case ID_DOT3_TANGENT_SPACE:
		g_bNormals = !g_bNormals;
		break;

	default:
		break;
	}

	return 0;
}

//=============================================================================
//-----------------------------------------------------------------------------
HRESULT CreateVS(const char * strVSName, LPDIRECT3DVERTEXSHADER9 * pVertexShader)
{
	// assemble and create the vertex shader
	DWORD dwFlags = 0;

#if defined( _DEBUG ) || defined( DEBUG )
	dwFlags |= D3DXSHADER_DEBUG;
#endif

	char			szFile[128];
	sprintf_s(szFile, "%s/%s", file_path, strVSName);

	char szBuffer[128];
	ID3DXBuffer*	pVS;

	HRESULT	rc;
	ID3DXBuffer*	pErrors;

	rc = D3DXAssembleShaderFromFile(szFile, NULL, NULL, dwFlags, &pVS, &pErrors);
	if (FAILED(rc))
	{
		OutputDebugString("Failed to assemble the vertex shader, errors:\n");
		OutputDebugString((char*)pErrors->GetBufferPointer());
		OutputDebugString("\n");
	}

	// Create the vertex shader
	rc = g_lpD3DDevice->CreateVertexShader((DWORD*)pVS->GetBufferPointer(), pVertexShader);
	if (FAILED(rc))
	{
		OutputDebugString("Failed to create the vertex shader, errors:\n");
		sprintf_s(szBuffer, _T("Error: %s error description: %s\n"), DXGetErrorString(rc), DXGetErrorDescription(rc));
		OutputDebugString(szBuffer);
		OutputDebugString("\n");
	}

	SAFE_RELEASE(pVS);
	if (FAILED(rc))
		return E_FAIL;

	return S_OK;
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
bool CreateVertexShader()
{
	HRESULT	rc;

	if (m_d3dCaps.VertexShaderVersion = D3DVS_VERSION(1, 1))
	{
		OutputDebugString(">> ... VertexShaderVersion 1.1 \n ");

		rc = CreateVS( SAMPLE_VS11, &g_lpVertexShader_11 );
		if ( S_OK == rc )
		{
			g_eVS_Start = eVS_1_1;
			g_eVS = eVS_1_1;
			g_eVS_MAX = eVS_1_1;
		}
	}
	 
	if ( m_d3dCaps.VertexShaderVersion = D3DVS_VERSION(2,0) )
	{
		OutputDebugString(">> ... VertexShaderVersion 2.0 \n ");

		rc = CreateVS( SAMPLE_VS20, &g_lpVertexShader_20 );

		if (S_OK == rc)
		{
			if (eVS_None == g_eVS)
			{
				g_eVS_Start = eVS_2_0;
				g_eVS = eVS_2_0;
				g_eVS_MAX = eVS_2_0;
			}
			else
			{
				g_eVS_MAX = eVS_2_0;
			}
		}
	}

	if ( eVS_None == g_eVS )
	{
		OutputDebugString(">> ... None VertexShader \n ");
		return false;
	}
	return true;
}

void FreeVertexShader()
{
	SAFE_RELEASE( g_lpVertexShader_11 );
	SAFE_RELEASE( g_lpVertexShader_20 );
}

//=============================================================================
//-----------------------------------------------------------------------------
HRESULT CreatePS( const char * strPSName, LPDIRECT3DPIXELSHADER9 * pPixelShader)
{
	// assemble and create the pixel shader
	DWORD dwFlags = 0;

#if defined( _DEBUG ) || defined( DEBUG )
	dwFlags |= D3DXSHADER_DEBUG;
#endif

	char			szFile[128];
	sprintf_s(szFile, "%s/%s", file_path, strPSName);

	char szBuffer[128];			// debug output
	ID3DXBuffer*	pPS;		// pointer to address space of the calling process

	HRESULT hr;					// error 
	ID3DXBuffer*	pErrors;

	hr = D3DXAssembleShaderFromFile(szFile, NULL, NULL, dwFlags, &pPS, &pErrors);
	if (FAILED(hr))
	{
		OutputDebugString("Failed to assemble the pixel shader, errors:\n");
		OutputDebugString((char*)pErrors->GetBufferPointer());
		OutputDebugString("\n");
	}

	// Create the pixel shader
	hr = g_lpD3DDevice->CreatePixelShader((DWORD*)pPS->GetBufferPointer(), pPixelShader);
	if (FAILED(hr))
	{
		OutputDebugString("Failed to create Pixel Shader, errors:\n");
		sprintf_s(szBuffer, _T("Error: %s error description: %s\n"), DXGetErrorString(hr), DXGetErrorDescription(hr));
		OutputDebugString(szBuffer);
		OutputDebugString("\n");
		return hr;
	}

	SAFE_RELEASE(pPS);
	if (FAILED(hr))
		return E_FAIL;

	return S_OK;
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
bool CreatePixelShader()
{
	HRESULT	rc;

	if (m_d3dCaps.PixelShaderVersion = D3DPS_VERSION(1, 1))
	{
		OutputDebugString(">> ... PixelShaderVersion 1.1 \n ");

		rc = CreatePS(SAMPLE_PS11, &g_lpPixelShader_11);
		if (S_OK == rc)
		{
			g_ePS_Start	= ePS_1_1;
			g_ePS		= ePS_1_1;
			g_ePS_MAX	= ePS_1_1;
		}
	}

	if (m_d3dCaps.PixelShaderVersion = D3DPS_VERSION(1, 4))
	{
		OutputDebugString(">> ... PixelShaderVersion 1.4 \n ");

		rc = CreatePS(SAMPLE_PS14, &g_lpPixelShader_14);
		if (S_OK == rc)
		{
			if (ePS_None == g_ePS)
			{
				g_ePS_Start = ePS_1_4;
				g_ePS = ePS_1_4;
				g_ePS_MAX = ePS_1_4;
			}
			else
			{
				g_ePS_MAX = ePS_1_4;
			}
		}
	}

	if (m_d3dCaps.PixelShaderVersion = D3DPS_VERSION(2, 0))
	{
		OutputDebugString(">> ... PixelShaderVersion 2.0 \n ");

		rc = CreatePS(SAMPLE_PS20, &g_lpPixelShader_20);
		if (S_OK == rc)
		{
			if (ePS_None == g_ePS)
			{
				g_ePS_Start = ePS_2_0;
				g_ePS = ePS_2_0;
				g_ePS_MAX = ePS_2_0;
			}
			else
			{
				g_ePS_MAX = ePS_2_0;
			}
		}
	}

	if (ePS_None == g_ePS)
	{
		OutputDebugString(">> ... None PixcelShader \n ");
		return false;
	}
	
	return true;
}

void FreePixelShader()
{
	SAFE_RELEASE( g_lpPixelShader_11 );
	SAFE_RELEASE( g_lpPixelShader_14 );
	SAFE_RELEASE( g_lpPixelShader_20 );
}

//=============================================================================
// Name: InitGeometry()
// Desc: Load the mesh and build the material and texture arrays
//-----------------------------------------------------------------------------
bool InitGeometry()
{
	HRESULT			hResult;
	LPD3DXBUFFER	lpD3DXMtrlBuffer;
	LPD3DXMESH		lpCloneMesh;

	// load mesh from x-file
	hResult = D3DXLoadMeshFromX(EARTH_MESH, D3DXMESH_SYSTEMMEM, g_lpD3DDevice, NULL,
		&lpD3DXMtrlBuffer, NULL, &g_dwMeshMaterials, &g_lpMesh);

	if (D3D_OK != hResult)
	{
		return false;
	}

	// convert mesh to appropriate vertex format
	hResult = g_lpMesh->CloneMeshFVF(0, FVF_VERTEX, g_lpD3DDevice, &lpCloneMesh);
	if (D3D_OK != hResult)
	{
		return false;
	}
	g_lpMesh = lpCloneMesh;

	// copy texture coordinates to second uv set
	// create tangent vectors for normal mapping
	LPDIRECT3DVERTEXBUFFER9		lpVB;
	LPDIRECT3DINDEXBUFFER9		lpIB;
	D3DVERTEXBUFFER_DESC		D3DVBDesc;
	D3DINDEXBUFFER_DESC			D3DIBDesc;
	DWORD						dwVertNum;
	DWORD						dwIndicesNum;
	SVertex*					pVertices;
	void*						pIndices;

	g_lpMesh->GetVertexBuffer(&lpVB);
	g_lpMesh->GetIndexBuffer(&lpIB);

	if ((NULL != lpVB) &&
		(NULL != lpIB))
	{
		// vertex buffer description
		lpVB->GetDesc(&D3DVBDesc);
		dwVertNum = D3DVBDesc.Size / sizeof(SVertex);
		g_nNormals = dwVertNum;

		// index buffer description
		lpIB->GetDesc(&D3DIBDesc);
		switch (D3DIBDesc.Format)
		{
		case D3DFMT_INDEX16:
			dwIndicesNum = D3DIBDesc.Size / sizeof(WORD);
			break;

		case D3DFMT_INDEX32:
			dwIndicesNum = D3DIBDesc.Size / sizeof(DWORD);
			break;
		}

		// lock mesh vertex buffer
		hResult = lpVB->Lock(0, 0, (void**)&pVertices, 0);
		if (D3D_OK != hResult)
		{
			return false;
		}

		// lock index buffer
		hResult = lpIB->Lock(0, 0, (void**)&pIndices, 0);
		if (D3D_OK != hResult)
		{
			return false;
		}

		// get data using appropriate tools
		std::vector<SVertexFix>		g_vertices;
		std::vector<int>			g_indices;

		// store vertex data
		DWORD i;
		for (i = 0; i < dwVertNum; i++)
		{
			SVertexFix	vertex;

			// prepare data for mesh correction
			vertex.vectPos = SVector3f(pVertices[i].x, pVertices[i].y, pVertices[i].z);
			vertex.vectUV = SVector2f(pVertices[i].u1, pVertices[i].v1);
			g_vertices.push_back(vertex);
		}

		// store index data
		for (i = 0; i < dwIndicesNum; i++)
		{
			int			nIndex;

			switch (D3DIBDesc.Format)
			{
			case D3DFMT_INDEX16:
			{
								   nIndex = ((WORD*)pIndices)[i];
								   g_indices.push_back(nIndex);
			}
				break;

			case D3DFMT_INDEX32:
			{
								   nIndex = ((DWORD*)pIndices)[i];
								   g_indices.push_back(nIndex);
			}
				break;
			}
		}

		// make calculations
		Dot3Mesh(g_vertices, g_indices);

		// restore tangent space
		for (i = 0; i < dwVertNum; i++)
		{
			pVertices[i].ux = g_vertices[i].vectTangent.x;
			pVertices[i].uy = g_vertices[i].vectTangent.y;
			pVertices[i].uz = g_vertices[i].vectTangent.z;

			pVertices[i].vx = g_vertices[i].vectBinormal.x;
			pVertices[i].vy = g_vertices[i].vectBinormal.y;
			pVertices[i].vz = g_vertices[i].vectBinormal.z;

			pVertices[i].uvx = g_vertices[i].vectNormal.x;
			pVertices[i].uvy = g_vertices[i].vectNormal.y;
			pVertices[i].uvz = g_vertices[i].vectNormal.z;
		}

		// --- fill buffers for tangent space ---
		int		nIndex = 0;

		g_pNormals = new SVertexVector[dwVertNum * 2];
		g_pTangents = new SVertexVector[dwVertNum * 2];
		g_pBinormals = new SVertexVector[dwVertNum * 2];

		for (i = 0; i < dwVertNum; i++)
		{
			// init normals
			g_pNormals[nIndex].pos.x = pVertices[i].x;
			g_pNormals[nIndex].pos.y = pVertices[i].y;
			g_pNormals[nIndex].pos.z = pVertices[i].z;

			g_pNormals[nIndex + 1].pos.x = g_pNormals[i * 2].pos.x + pVertices[i].uvx * 0.10f;
			g_pNormals[nIndex + 1].pos.y = g_pNormals[i * 2].pos.y + pVertices[i].uvy * 0.10f;
			g_pNormals[nIndex + 1].pos.z = g_pNormals[i * 2].pos.z + pVertices[i].uvz * 0.10f;

			g_pNormals[nIndex].color = 0xffffff00;
			g_pNormals[nIndex + 1].color = 0xffffff00;

			// init tangents
			g_pTangents[nIndex].pos.x = pVertices[i].x;
			g_pTangents[nIndex].pos.y = pVertices[i].y;
			g_pTangents[nIndex].pos.z = pVertices[i].z;

			g_pTangents[nIndex + 1].pos.x = g_pTangents[i * 2].pos.x + pVertices[i].ux * 0.10f;
			g_pTangents[nIndex + 1].pos.y = g_pTangents[i * 2].pos.y + pVertices[i].uy * 0.10f;
			g_pTangents[nIndex + 1].pos.z = g_pTangents[i * 2].pos.z + pVertices[i].uz * 0.10f;

			g_pTangents[nIndex].color = 0xffff0000;
			g_pTangents[nIndex + 1].color = 0xffff0000;

			// init binormals
			g_pBinormals[nIndex].pos.x = pVertices[i].x;
			g_pBinormals[nIndex].pos.y = pVertices[i].y;
			g_pBinormals[nIndex].pos.z = pVertices[i].z;

			g_pBinormals[nIndex + 1].pos.x = g_pBinormals[i * 2].pos.x + pVertices[i].vx * 0.10f;
			g_pBinormals[nIndex + 1].pos.y = g_pBinormals[i * 2].pos.y + pVertices[i].vy * 0.10f;
			g_pBinormals[nIndex + 1].pos.z = g_pBinormals[i * 2].pos.z + pVertices[i].vz * 0.10f;

			g_pBinormals[nIndex].color = 0xff00ff00;
			g_pBinormals[nIndex + 1].color = 0xff00ff00;

			nIndex += 2;
		}

		// --- end of tangent space ---

		// unlock mesh vertex buffer
		hResult = lpVB->Unlock();
		if (D3D_OK != hResult)
		{
			return false;
		}

		// unlock mesh index buffer
		hResult = lpIB->Unlock();
		if (D3D_OK != hResult)
		{
			return false;
		}
	}

	// release vertex and index buffer
	SAFE_RELEASE(lpIB);
	SAFE_RELEASE(lpVB);

	// load mesh textures
	D3DXMATERIAL*	D3DXMaterials = (D3DXMATERIAL*)lpD3DXMtrlBuffer->GetBufferPointer();

	g_lpMeshTextures = new LPDIRECT3DTEXTURE9[g_dwMeshMaterials];
	assert(g_lpMeshTextures);

	DWORD i;
	for (i = 0; i < g_dwMeshMaterials; i++)
	{
		char szBuffer[MAX_PATH];

		// prepare buffer
		memset(szBuffer, 0, MAX_PATH);

		// store path if needed
		if (NULL != D3DXMaterials[i].pTextureFilename)
		{
			sprintf_s(szBuffer, "Media/%s", D3DXMaterials[i].pTextureFilename);
		}

		// load object texture
		hResult = D3DXCreateTextureFromFile(g_lpD3DDevice, szBuffer, &g_lpMeshTextures[i]);
		if (D3D_OK != hResult)
		{
			g_lpMeshTextures[i] = NULL;
		}
	}
	SAFE_RELEASE(lpD3DXMtrlBuffer);

	// load normal texture
	hResult = D3DXCreateTextureFromFile(g_lpD3DDevice, EARTH_NORMAL_TEX, &g_lpMeshNormalTexture);
	if (D3D_OK != hResult)
	{
		g_lpMeshTextures[i] = NULL;
	}

	// load white texture
	hResult = D3DXCreateTextureFromFile(g_lpD3DDevice, WHITE_TEX, &g_lpWhiteTexture);
	if (D3D_OK != hResult)
	{
		g_lpWhiteTexture = NULL;
	}

	// load cube enviromental texture
	hResult = D3DXCreateCubeTextureFromFile(g_lpD3DDevice, ENV_CUBE_MAP, &g_lpCubeEnvTexture);
	if (D3D_OK != hResult)
	{
		return false;
	}

	LPDIRECT3DSURFACE9 pFace2;
	g_lpCubeEnvTexture->GetCubeMapSurface(D3DCUBEMAP_FACE_POSITIVE_Y, 0, &pFace2);

	return true;
}

void FreeGeometry()
{
	// free tangent space
	SAFE_DELETE_ARRAY(g_pNormals);
	SAFE_DELETE_ARRAY(g_pTangents);
	SAFE_DELETE_ARRAY(g_pBinormals);

	// free textures
	for (DWORD i = 0; i < g_dwMeshMaterials; i++)
	{
		SAFE_RELEASE(g_lpMeshTextures[i]);
	}
	SAFE_DELETE_ARRAY(g_lpMeshTextures);
	SAFE_RELEASE(g_lpMeshNormalTexture);
	SAFE_RELEASE(g_lpCubeEnvTexture);

	// free mesh
	SAFE_RELEASE(g_lpMesh);
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
