﻿////////////////////////////////////////////////////////////////////////////////
// RacorX52 재구성 
//
// 참고 ; Book_2004_ShaderX²_DirectX 9 셰이더 프로그래밍_CD\03 - Michal Valient\ Source\ Utils\ Tester02
// 2017.04.20.(목)
//
//=============================================================================
// RacorX5 재구성 
//-----------------------------------------------------------------------------
// RacorX4에 점광원을 추가 (감쇠 모형을 가진 점광원)
// <Home> 키와 <End> 키, 그리고 상하좌우 화살표 키들로 그 광원의 위치를 이동
// <C> 키로 반영광의 색을 변경하는 것도 가능
// 2016.08.22.(월)
//
//=============================================================================
// RacorX4 
//-----------------------------------------------------------------------------
// 1. 베지어 패치를 이용한 곡면을 표시 + 텍스처 설정 + 분산과 반영이(specular)이 결합된 조명
// 2. 객체 회전, 마우스 휠로 확대, 축소
// 3. <C> 키로 반영 색상을 선택
// 2016.08.19.(금)
//
//=============================================================================
// RacorX3
//-----------------------------------------------------------------------------
// 1. 정점 당 분산 반사(diffuse reflection) 모형이 추가
// 2. SetFVF( ... ); 을 사용하지 않고 SetVertexDeclaration( ...); 함수를 사용하도록 수정
//		--> LPDIRECT3DVERTEXDECLARATION9	m_pVertexDeclaration; 설정
// 3. 상수 레지스터 이름들이 헤더 파일 const.h에 저장
// 2016.08.18.(목)
// 
//=============================================================================
// RacorX202 재구성 
//-----------------------------------------------------------------------------
// Debug 모드 ; "Shader\\basic.vsh"을 읽어 와서 "Shader\\basic.bin"을 생성 후 shader 실행
// Release 모드 ; "Shader\\basic.bin"를 읽어 와서 shader 실행
// 2016.08.17.(수)
//
//=============================================================================
// RacorX201 재구성 
//-----------------------------------------------------------------------------
// Shader 정의를 외부 파일에서 읽어 옴 ; D3DXAssembleShader(); --> D3DXAssembleShaderFromFile("basic.vsh", ... );
// 2016.08.16.(화)
//
//=============================================================================
// RacorX102
// Direct3D ShaderX 책의 샘플 환경 ; DirectX 8, Visual Studio 6.0)
//-----------------------------------------------------------------------------
// DirectX 9.0c, Visual Studio 2013 환경에서 빌드
// DirectX 9.0b의 commom 폴더 파일 ; DirectX 9.0b의 공용프레임웍 파일
// 2016.08.12.(금)
//
////////////////////////////////////////////////////////////////////////////////


#include "_StdAfx.h"
#include <fcntl.h>		// O_RDONLY
#include <io.h>			// _open

const char * file_path		= "Shader";
const char * file_vsh		= "52_v_BumpReflection";
const char * file_psh		= "52_p_BumpReflection";

const char * file_normal	= "media/earth_normal.tga";
const char * file_cubeEnv	= "media/LobbyCube.dds";
const char * file_duffuse	= "media/ShaderX.tga";


// Create the shader declaration.
D3DVERTEXELEMENT9 decl[] =
{
	{ 0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
	{ 0, 12, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL, 0 },
	{ 0, 24, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },
	{ 0, 32, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT, 0 },
	D3DDECL_END()
};

//=============================================================================
//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the 
//       generic functionality needed in all Direct3D samples. CMyD3DApplication 
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
protected:
	CD3DFont*		m_pFont;				// Font for drawing tex
	CD3DBPatch*		m_pBPatch;

	D3DXMATRIX		m_matProj, m_matView, m_matWorld, m_matWorldI;

	D3DXVECTOR3		m_vObjectCenter;		// Center of bounding sphere of object
	FLOAT			m_fObjectRadius;		// Radius of bounding sphere of object

	CD3DArcBall		m_ArcBall;				// trackball rotation utility
	FLOAT			m_fZDist;

	D3DXVECTOR4		m_vLight;				// light position or direction

	LPDIRECT3DVERTEXSHADER9			m_pVertexShader;	// vertex shader handle
	LPDIRECT3DVERTEXDECLARATION9	m_pVertexDeclaration;

	LPDIRECT3DPIXELSHADER9			m_pPixelShader;		// pixel shader handle
	
	TCHAR							m_cNormal[64];
	LPDIRECT3DTEXTURE9				m_pNormalTexture;
	TCHAR							m_cCubeEnv[64];
	LPDIRECT3DCUBETEXTURE9			m_pCubeEnvTexture;
	TCHAR							m_cDiffuse[64];
	LPDIRECT3DTEXTURE9				m_pDiffuseTexture;
	
	BYTE m_bKey[256];					// keyboard state buffer
	
	HRESULT ConfirmDevice( D3DCAPS9*, DWORD, D3DFORMAT );

	HRESULT OneTimeSceneInit();
	HRESULT InitDeviceObjects();
	HRESULT RestoreDeviceObjects();
	HRESULT InvalidateDeviceObjects();
	HRESULT DeleteDeviceObjects();
	HRESULT FinalCleanup();
	HRESULT Render();
	HRESULT FrameMove();
	LRESULT MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

public:
	CMyD3DApplication();

private:
	HRESULT CreateVS(const char * strVSName, LPDIRECT3DVERTEXSHADER9 * pVertexShader);
	HRESULT CreatePS(const char * strPSName, LPDIRECT3DPIXELSHADER9 * pPixelShader);

};


//=============================================================================
//-----------------------------------------------------------------------------
CMyD3DApplication g_d3dApp;

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
	if( FAILED( g_d3dApp.Create( hInst ) ) )
		return 0;

	return g_d3dApp.Run();
}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
{
	m_strWindowTitle	= _T("RacorX #5");
	m_dwCreationWidth	= 800;	// Width used to create window
	m_dwCreationHeight	= 600;
	m_bUseDepthBuffer	= TRUE;
	m_pFont = new CD3DFont(_T("Arial"), 12, D3DFONT_BOLD);
	m_pVertexShader = NULL;
	m_pVertexDeclaration = NULL;

	m_pPixelShader = NULL;

	m_pNormalTexture	= NULL;
	m_pCubeEnvTexture	= NULL;
	m_pDiffuseTexture	= NULL;

	m_pBPatch = new CD3DBPatch();
	m_strWindowTitle = "52_Shader_BumpReflection";
	m_fObjectRadius		= 1;
	ZeroMemory(m_bKey, 256);
	m_vLight = D3DXVECTOR4(0, 0, -1, 0.4);		// Point Light Position

}

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
	strcpy( m_cNormal, _T(file_normal ) );
	strcpy( m_cCubeEnv, _T(file_cubeEnv ) );
	strcpy( m_cDiffuse, _T( file_duffuse ) );
	
	return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
	// rotates and translate the object
	D3DXMatrixTranslation(&m_matWorld, -m_vObjectCenter.x, -m_vObjectCenter.y,	-m_vObjectCenter.z);

	D3DXMatrixMultiply(&m_matWorld, &m_matWorld, m_ArcBall.GetTranslationMatrix());
	D3DXMatrixMultiply(&m_matWorld, &m_matWorld, m_ArcBall.GetRotationMatrix());

	//Set up the view matrix.
	D3DXVECTOR3 vEyePt( 0.0f, 0.0f, -3.7f * m_fObjectRadius + m_fZDist );
	D3DXVECTOR3 vLookAtPt( 0.0f, 0.0f, 0.0f );
	D3DXVECTOR3 vUp( 0.0f, 1.0f, 0.0f );
	D3DXMatrixLookAtLH( &m_matView, &vEyePt, &vLookAtPt, &vUp );

	// shaders data
	D3DXMATRIX matTemp;
	D3DXMatrixTranspose(&matTemp, &(m_matWorld * m_matView * m_matProj));

	m_pd3dDevice->SetVertexShaderConstantF(0, (float*)&m_matWorld, 4);
	m_pd3dDevice->SetVertexShaderConstantF(4, (float*)&matTemp, 4);
		
	// set eye vector E
	m_pd3dDevice->SetVertexShaderConstantF(8, vEyePt, 1);

	//Vertex addition factor for cube map
	LPDIRECT3DSURFACE9 pFace2;
	m_pCubeEnvTexture->GetCubeMapSurface(D3DCUBEMAP_FACE_POSITIVE_Y, 0, &pFace2);

	float Size = 256.0f;
	D3DXVECTOR4 VertFactor(1.0f / Size, 1.0f, 1.0f, 1.0f);
	m_pd3dDevice->SetVertexShaderConstantF(9, VertFactor, 1);

	if (m_bKey[VK_LEFT])	{ m_bKey[VK_RIGHT] = 0; m_vLight.x -= 1.0f * m_fElapsedTime; }
	if (m_bKey[VK_RIGHT])	{ m_bKey[VK_LEFT] = 0;	m_vLight.x += 1.0f * m_fElapsedTime; }
	if (m_bKey[VK_UP])		{ m_bKey[VK_DOWN] = 0;	m_vLight.y += 1.0f * m_fElapsedTime; }
	if (m_bKey[VK_DOWN])	{ m_bKey[VK_UP] = 0;	m_vLight.y -= 1.0f * m_fElapsedTime; }
	if (m_bKey[VK_END])		{ m_bKey[VK_HOME] = 0;	m_vLight.z += 1.0f * m_fElapsedTime; }
	if (m_bKey[VK_HOME])	{ m_bKey[VK_END] = 0;	m_vLight.z -= 1.0f * m_fElapsedTime; }

/*
	Refraction indexes. Taken from Matthias Wloka paper at www.nvidia.com.

	Material                Index of Refraction
	Vacuum                  1.0 (min. index of refraction)
	Air                     1.000293
	Ice                     1.31
	Water                   1.333333
	Ethyl Alcohol           1.36
	Fluorite                1.43
	Poppy Seed Oil          1.469
	Olive Oil               1.47
	Linseed Oil             1.478
	Plexiglas               1.51
	Immersion Oil           1.515
	Crown Glass             1.52
	Quartz                  1.54
	Salt                    1.54
	Light Flint Glass       1.58
	Dense Flint Glass       1.66
	Tourmaline              1.62
	Garnet                  1.73-1.89
	Zircon                  1.923
	Cubic Zirconia          2.14-2.20
	Diamond                 2.417
	Rutile                  2.907
	Gallium Phosphide       3.5 (max. index of refraction)
*/

	//precomputed refraction index
	float refractionIndexRatio = 3.5f;
	//float refractionIndexRatio = 1.0f;
	float rIndex = powf(1.0f - refractionIndexRatio, 2.0f) / powf(1.0f + refractionIndexRatio, 2.0f);
	//rIndex = 1;
	D3DXCOLOR RefIndex(rIndex, 1.0, 1.0, 1.0);
	m_pd3dDevice->SetPixelShaderConstantF(0, RefIndex, 1);

	return S_OK;
}

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::Render()
{
	// Clear the viewport | z-buffer
	m_pd3dDevice->Clear(0L, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
		D3DCOLOR_XRGB(0, 0, 128), 1.0f, 0L);

	// Begin the scene
	if (SUCCEEDED(m_pd3dDevice->BeginScene()))
	{
		// set shader
		m_pd3dDevice->SetVertexShader(m_pVertexShader);
		m_pd3dDevice->SetPixelShader(m_pPixelShader);

		// set FVF
		// m_pd3dDevice->SetFVF(D3DFVF_VERTEX);
		m_pd3dDevice->SetVertexDeclaration(m_pVertexDeclaration);

		m_pd3dDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
		m_pd3dDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
		m_pd3dDevice->SetSamplerState(3, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
		m_pd3dDevice->SetSamplerState(3, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);

		m_pd3dDevice->SetSamplerState(3, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP);
		m_pd3dDevice->SetSamplerState(3, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP);

		// set the texture
		m_pd3dDevice->SetTexture(0, m_pDiffuseTexture);
		m_pd3dDevice->SetTexture(1, m_pNormalTexture);
		m_pd3dDevice->SetTexture(2, m_pCubeEnvTexture);
		
		m_pd3dDevice->SetTextureStageState(3, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_CAMERASPACENORMAL | 1);

		m_pBPatch->Render(m_pd3dDevice);
		
		m_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);

		// Output statistics
		m_pFont->DrawText(2, 0, D3DCOLOR_ARGB(255, 255, 255, 0), m_strFrameStats);
		m_pFont->DrawText(2, 20, D3DCOLOR_ARGB(255, 255, 255, 0), m_strDeviceStats);

		// End the scene.
		m_pd3dDevice->EndScene();
	}

	return S_OK;
}

//------------------------------------------------------------------------------
// Name: CreateVS()
// Desc:
//------------------------------------------------------------------------------
HRESULT CMyD3DApplication::CreateVS(
	const char * strVSName, LPDIRECT3DVERTEXSHADER9 * pVertexShader)
{
	// assemble and create the vertex shader
	DWORD dwFlags = 0;

#if defined( _DEBUG ) || defined( DEBUG )
	dwFlags |= D3DXSHADER_DEBUG;
#endif

	char			szFile[128];
	char			szFileBin[128];
	sprintf(szFile, "%s/%s.vsh", file_path, strVSName);
	sprintf(szFileBin, "%s/%s.bin", file_path, strVSName);

	char szBuffer[128];
	ID3DXBuffer*	pVS;

	HRESULT	rc;

#if defined( _DEBUG ) || defined( DEBUG )
	ID3DXBuffer*	pErrors;

	rc = D3DXAssembleShaderFromFile(szFile, NULL, NULL, dwFlags, &pVS, &pErrors);
	if (FAILED(rc))
	{
		OutputDebugString("Failed to assemble the vertex shader, errors:\n");
		OutputDebugString((char*)pErrors->GetBufferPointer());
		OutputDebugString("\n");
	}
	else
	{
		FILE* fp = NULL;
		fp = fopen(szFileBin, "wb");
		if (fp != NULL){
			fwrite(pVS->GetBufferPointer(), pVS->GetBufferSize(), 1, fp);
			fclose(fp);
		}
	}
#else
	int handle = _open(szFileBin, O_RDONLY | O_BINARY, NULL);
	if (handle == -1){
		OutputDebugString("Read VertexShader Error.. \r\n");
		return E_FAIL;
	}

	int nLength = _filelength(handle);
	D3DXCreateBuffer(nLength, &pVS);
	_read(handle, pVS->GetBufferPointer(), nLength);

#endif

	// Create the vertex shader
	rc = m_pd3dDevice->CreateVertexShader((DWORD*)pVS->GetBufferPointer(), pVertexShader);
	if (FAILED(rc))
	{
		OutputDebugString("Failed to create the vertex shader, errors:\n");
		_stprintf(szBuffer, _T("Error: %s error description: %s\n"), DXGetErrorString(rc), DXGetErrorDescription(rc));
		OutputDebugString(szBuffer);
		OutputDebugString("\n");
	}

	SAFE_RELEASE(pVS);
	if (FAILED(rc))
		return E_FAIL;
	
	return S_OK;
}

//------------------------------------------------------------------------------
// Name: CreatePS()
// Desc:
//------------------------------------------------------------------------------
HRESULT CMyD3DApplication::CreatePS(
	const char * strPSName, LPDIRECT3DPIXELSHADER9 * pPixelShader)
{
	// assemble and create the pixel shader
	DWORD dwFlags = 0;

#if defined( _DEBUG ) || defined( DEBUG )
	dwFlags |= D3DXSHADER_DEBUG;
#endif

	char			szFile[128];
	char			szFileBin[128];
	sprintf(szFile, "%s/%s.psh", file_path, strPSName);
	sprintf(szFileBin, "%s/%s.bin", file_path, strPSName);

	char szBuffer[128];			// debug output
	ID3DXBuffer*	pPS;		// pointer to address space of the calling process

	HRESULT hr;					// error 
#if defined( _DEBUG ) || defined( DEBUG )
	ID3DXBuffer*	pErrors;

	hr = D3DXAssembleShaderFromFile(szFile, NULL, NULL, dwFlags, &pPS, &pErrors);
	if (FAILED(hr))
	{
		OutputDebugString("Failed to assemble the pixel shader, errors:\n");
		OutputDebugString((char*)pErrors->GetBufferPointer());
		OutputDebugString("\n");
	}
	else
	{
		FILE* fp = NULL;
		fp = fopen(szFileBin, "wb");
		if (fp != NULL){
			fwrite(pPS->GetBufferPointer(), pPS->GetBufferSize(), 1, fp);
			fclose(fp);
		}
	}
#else
	int handle = _open(szFileBin, O_RDONLY | O_BINARY, NULL);
	if (handle == -1){
		OutputDebugString("Read VertexShader Error.. \r\n");
		return E_FAIL;
	}

	int nLength = _filelength(handle);
	D3DXCreateBuffer(nLength, &pPS);
	_read(handle, pPS->GetBufferPointer(), nLength);

#endif

	// Create the pixel shader
	hr = m_pd3dDevice->CreatePixelShader((DWORD*)pPS->GetBufferPointer(), pPixelShader);
	if (FAILED(hr))
	{
		OutputDebugString("Failed to create Pixel Shader, errors:\n");
		_stprintf(szBuffer, _T("Error: %s error description: %s\n"), DXGetErrorString(hr), DXGetErrorDescription(hr));
		OutputDebugString(szBuffer);
		OutputDebugString("\n");
		return hr;
	}

	SAFE_RELEASE(pPS);
	if (FAILED(hr))
		return E_FAIL;

	return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: This creates all device-dependent managed objects, such as managed
//       textures and managed vertex buffers.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
	// Initialize the font's internal textures
	m_pFont->InitDeviceObjects(m_pd3dDevice);

	// Load the texture
	HRESULT	hResult;

	hResult = D3DXCreateTextureFromFile( m_pd3dDevice, m_cDiffuse, &m_pDiffuseTexture );
	if ( S_OK != hResult )
	{
		return S_FALSE;
	}

	// Load bump map or normal map from height map
	LPDIRECT3DTEXTURE9 pHeightMap = NULL;
	hResult = D3DXCreateTextureFromFile(m_pd3dDevice, m_cNormal, &pHeightMap);
	if ( S_OK != hResult )
	{
		return S_FALSE;
	}

	D3DSURFACE_DESC    desc;
	pHeightMap->GetLevelDesc(0, &desc);
	if (FAILED(D3DXCreateTexture(m_pd3dDevice, desc.Width, desc.Height, 0, 0,
		D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &m_pNormalTexture)))
		return E_FAIL;

	D3DXComputeNormalMap(m_pNormalTexture, pHeightMap, NULL, 0, D3DX_CHANNEL_RED, 10);
	SAFE_RELEASE(pHeightMap);
	
	// load cube enviromental texture
	hResult = D3DXCreateCubeTextureFromFile(m_pd3dDevice, m_cCubeEnv, &m_pCubeEnvTexture );
	if (S_OK != hResult)
	{
		return S_FALSE;
	}

	LPDIRECT3DSURFACE9 pFace2;
	m_pCubeEnvTexture->GetCubeMapSurface(D3DCUBEMAP_FACE_POSITIVE_Y, 0, &pFace2);

	VERTICES cp[9];
	cp[0].vPosition = D3DXVECTOR3(-1, 1, 0);	 cp[0].uv = D3DXVECTOR2(0.0f, 0.0f);
	cp[1].vPosition = D3DXVECTOR3(0, 1, 0);      cp[1].uv = D3DXVECTOR2(0.5f, 0.0f);
	cp[2].vPosition = D3DXVECTOR3(1, 1, 0);      cp[2].uv = D3DXVECTOR2(1.0f, 0.0f);

	cp[3].vPosition = D3DXVECTOR3(-1, 0, 0);     cp[3].uv = D3DXVECTOR2(0.0f, 0.5f);
	cp[4].vPosition = D3DXVECTOR3(0, 0, -2);     cp[4].uv = D3DXVECTOR2(0.5f, 0.5f);
	cp[5].vPosition = D3DXVECTOR3(1, 0, 0);      cp[5].uv = D3DXVECTOR2(1.0f, 0.5f);

	cp[6].vPosition = D3DXVECTOR3(-1, -1, 0);    cp[6].uv = D3DXVECTOR2(0.0f, 1.0f);
	cp[7].vPosition = D3DXVECTOR3(0, -1, 0);     cp[7].uv = D3DXVECTOR2(0.5f, 1.0f);
	cp[8].vPosition = D3DXVECTOR3(1, -1, 0);     cp[8].uv = D3DXVECTOR2(1.0f, 1.0f);
	
	m_pBPatch->m_dwULevel = 4;
	m_pBPatch->m_dwVLevel = 4;
	m_pBPatch->Create(cp, 3, 3);

	m_pd3dDevice->CreateVertexDeclaration( decl, &m_pVertexDeclaration );

	CreateVS( file_vsh, &m_pVertexShader );
	CreatePS( file_psh, &m_pPixelShader );

	return S_OK;
}

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
	// Restore the device objects for the font
	m_pFont->RestoreDeviceObjects();
	m_pBPatch->RestoreDeviceObjects(m_pd3dDevice);

	// Set the projection matrix
	FLOAT fAspect = m_d3dsdBackBuffer.Width / (FLOAT)m_d3dsdBackBuffer.Height;
	D3DXMatrixPerspectiveFovLH( &m_matProj, D3DX_PI / 4, fAspect, m_fObjectRadius / 64.0f, m_fObjectRadius*200.0f );

	// set arc ball
	m_ArcBall.SetWindow(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 0.85f);
	m_ArcBall.SetTranslationRadius( m_fObjectRadius );

	// dithering on
	m_pd3dDevice->SetRenderState(D3DRS_DITHERENABLE, TRUE);
	
	// z-buffer enabled
	m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, TRUE);

	// Turn off D3D lighting, since we are providing our own vertex shader lighting
	m_pd3dDevice->SetRenderState(D3DRS_LIGHTING, FALSE);

	// Turn off culling, so we see the front and back of the quad
	m_pd3dDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
	
	// enable specular lighting
	m_pd3dDevice->SetRenderState(D3DRS_SPECULARENABLE, TRUE);

	return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
	m_pFont->InvalidateDeviceObjects();
	m_pBPatch->InvalidateDeviceObjects();

	SAFE_RELEASE(m_pVertexDeclaration);

	return S_OK;
}

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
	m_pFont->DeleteDeviceObjects();
	m_pBPatch->DeleteDeviceObjects();

	SAFE_RELEASE( m_pNormalTexture );
	SAFE_RELEASE( m_pCubeEnvTexture );
	SAFE_RELEASE( m_pDiffuseTexture );
	
	SAFE_RELEASE( m_pVertexShader );
	SAFE_RELEASE( m_pPixelShader );

	return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
	SAFE_DELETE(m_pFont);
	SAFE_DELETE(m_pBPatch);

	return S_OK;
}

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice(D3DCAPS9* pCaps, DWORD dwBehavior,	D3DFORMAT Format)
{
	// 정점 셰이더 지원 여부의 확인
	if ((dwBehavior & D3DCREATE_HARDWARE_VERTEXPROCESSING) ||
		(dwBehavior & D3DCREATE_MIXED_VERTEXPROCESSING))
	{
		if (pCaps->VertexShaderVersion < D3DVS_VERSION(1, 1))
			return E_FAIL;
	}

	return S_OK;
}

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: 
//
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	// Pass mouse messages to the ArcBall so it can build internal matrices
	m_ArcBall.HandleMessages( hWnd, msg, wParam, lParam );

	switch (msg)
	{
	case WM_KEYDOWN:
		m_bKey[wParam] = 1;
		break;

	case WM_KEYUP:
		m_bKey[wParam] = 0;
		break;

	case WM_MOUSEWHEEL:
		short sr = HIWORD(wParam);
		m_fZDist -= (sr / 120) * m_fObjectRadius;
		break;

	}

	return CD3DApplication::MsgProc(hWnd, msg, wParam, lParam);
}
